/*
	SerLCD controller, based on SparkFun's version! 
	RobThePyro - 28/4/2012
	Rambo.co.za!
	To all the authors of the some of the code I have used below, I hope you don't mind that I adapted it for my own needs... Please contact me if you have any issues! :D 

    LaunchPad                                    LCD


    TP 1  Outputs 5V from USB -   Pin 2 and Pin15 5V Vcc

    Vcc- nc
    P1.0 -                        Pin 4 RSk
    P1.1 - PC UART interface
    P1.2 - PC UART interface
    P1.3 -                        Pin 6 Enable
    P1.4 -                        Pin 11 DB4
    P1.5 -                        Pin 12 DB5
    P1.6 -                        Pin 13 DB6
    P1.7 -                        Pin 14 DB7
    GND -                         Pin 5 R/W, Pin 1 and Pin 16 GND
                                  Pin 3 Resistor* 1800 ohm to GND
*/

	#include <msp430.h>
    #include "stdint.h"

    //------------------------------------------------------------------------------
    // Hardware-related definitions
    //------------------------------------------------------------------------------
    #define UART_TXD   0x02                     // TXD on P1.1 (Timer0_A.OUT0)
    #define UART_RXD   0x04                     // RXD on P1.2 (Timer0_A.CCI1A)

    //------------------------------------------------------------------------------
    // Conditions for 9600 Baud SW UART, SMCLK = 1MHz
    //------------------------------------------------------------------------------
    #define UART_TBIT_DIV_2     (1000000 / (9600 * 2))
    #define UART_TBIT           (1000000 / 9600)

    //------------------------------------------------------------------------------
    // Global variables used for full-duplex UART communication
    //------------------------------------------------------------------------------
    unsigned int txData;                        // UART internal variable for TX
    unsigned char rxBuffer;                     // Received UART character
    unsigned char BLBright;
    uint16_t CommandFlag;
    uint16_t SecondFlag;
    #define MAGIC 0xFE // command flag <<
	#define Control	0x7c


    //------------------------------------------------------------------------------
    //LCD
    //------------------------------------------------------------------------------
    #define LCD_MODE_BIT 0x01   //BIT0
    #define LCD_ENABLE_BIT 0x08   //BIT3

    #define LCD_CMD_MASK 0x09
    #define LCD_DATA_MASK 0xf0
    #define LCD_PORT_OUT P1OUT
    #define LCD_PORT_DIR P1DIR



    int putchar( uint16_t in )
    {
       LCD_PORT_OUT= ( in & 0xf0 ) | LCD_MODE_BIT;
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       //
       LCD_PORT_OUT= ( ( in << 4 ) & 0xf0 ) | LCD_MODE_BIT;
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       __no_operation();
       //LCD_PORT_OUT= 0;
       __delay_cycles( 200 );
       return (0);
    }

    void lcd_command( uint16_t cmd )
    {
       LCD_PORT_OUT= ( cmd & 0xf0 );
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       //
       LCD_PORT_OUT= ( ( cmd << 4 ) & 0xf0 );
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       __no_operation();
       //LCD_PORT_OUT= 0;
       __delay_cycles( 3000 );
    }



    //------------------------------------------------------------------------------
    // Function prototypes
    //------------------------------------------------------------------------------
    void setBL(void);
    void TimerA_UART_init(void);
    void TimerA_UART_tx(unsigned char byte);
    void TimerA_UART_print(char *string);

    //------------------------------------------------------------------------------
    // main()
    //------------------------------------------------------------------------------
    void main(void)
    {
        WDTCTL = WDTPW + WDTHOLD;               // Stop watchdog timer

        DCOCTL = 0x00;                          // Set DCOCLK to 1MHz
        BCSCTL1 = CALBC1_1MHZ;
        DCOCTL = CALDCO_1MHZ;

        //P1OUT = 0x00;                           // Initialize all GPIO
        P1SEL = UART_TXD + UART_RXD;            // Timer function for TXD/RXD pins
        P1DIR = 0xFF & ~UART_RXD;               // Set all pins but RXD to output
        //P2OUT = 0x00;
        P2SEL = 0x00; // use Xin Xout pins
        P2DIR = 0xFF;

       //lcd
       //LCD_PORT_OUT= 0;
       LCD_PORT_DIR|= LCD_DATA_MASK | LCD_CMD_MASK;

       //don't assume VCC+ > 20ms away
       __delay_cycles(1000000);
       //first init
       LCD_PORT_OUT= 0x30;
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       //pause should be set by mclk div, > 4.1ms
       __delay_cycles( 15000 );

       //second init
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       //pause should be set by mclk div, > 100us
       __delay_cycles( 500 );

       //four bit mode
       LCD_PORT_OUT= 0x20;
       LCD_PORT_OUT|= LCD_ENABLE_BIT;
       __no_operation();
       LCD_PORT_OUT&= ~LCD_ENABLE_BIT;
       __delay_cycles( 1000 );

       //from here 'lcd_command can be used
       //operational mode 5x8 1/16 duty( two lines )
       lcd_command( 0x28 );
       __delay_cycles( 1000 );

       //display off
       lcd_command( 0x08 );
       __delay_cycles( 1000 );

       //display on, no cursor, no blink...
       lcd_command( 0x0c );
       __delay_cycles( 1000 );
       //LCD_PORT_OUT= 0;


       // Clear the display
       lcd_command(0x01);

       // splashy screen! :D

       putchar ('/');
       putchar ('/');
       putchar (' ');
       putchar ('S');
       putchar ('e');
       putchar ('r');
       putchar ('L');
       putchar ('C');
       putchar ('D');
       putchar ('!');
       putchar (' ');
       putchar ('V');
       putchar ('0');
       putchar ('.');
       putchar ('8');
       lcd_command(0xC0); // goto next line
       putchar (' ');
       putchar (' ');
       putchar ('>');
       putchar ('R');
       putchar ('A');
       putchar ('M');
       putchar ('B');
       putchar ('O');
       putchar ('.');
       putchar ('c');
       putchar ('o');
       putchar ('.');
       putchar ('z');
       putchar ('a');
       putchar ('<');





        __enable_interrupt();

        TimerA_UART_init();                     // Start Timer_A UART
        TimerA_UART_print("RobThePyro's MSP_430_serLCD adaption! \r\n");
        TimerA_UART_print("READY. :D\r\n");
        //for (;;)TimerA_UART_print("12345678901234567890123456789012345678901234567890123456789012345678901234567890");

        for (;;)
        {
            // Wait for incoming character
            __bis_SR_register(LPM0_bits);

            if(!CommandFlag && !SecondFlag){ // not a command waiting
            	if(rxBuffer !=MAGIC && rxBuffer !=Control ){ // put char to screen
            		putchar(rxBuffer);
            	} else if(rxBuffer == MAGIC){ // LCD control char
            		CommandFlag=1;
            	} else { // Device control char
            		SecondFlag=1;
            	}
            }else if(CommandFlag){ // lcd command waiting
            	CommandFlag=0;
            	lcd_command(rxBuffer);
            }else if(SecondFlag){ // device command waiting
            	SecondFlag = 0;
            	BLBright = rxBuffer;
            	setBL();
            }

            // Echo received character
            //TimerA_UART_tx(rxBuffer);
        }
    }
    //------------------------------------------------------------------------------
    // Function configures the BackLight
    //------------------------------------------------------------------------------
    void setBL(void){
    	if(BLBright == 0x80){ // led off
    		P2OUT &= ~0x80;
    	}else{ //led on
    		P2OUT |= 0x80;
    	}
    }

    //------------------------------------------------------------------------------
    // Function configures Timer_A for full-duplex UART operation
    //------------------------------------------------------------------------------
    void TimerA_UART_init(void)
    {
        TACCTL0 = OUT;                          // Set TXD Idle as Mark = '1'
        TACCTL1 = SCS + CM1 + CAP + CCIE;       // Sync, Neg Edge, Capture, Int
        TACTL = TASSEL_2 + MC_2;                // SMCLK, start in continuous mode
    }
    //------------------------------------------------------------------------------
    // Outputs one byte using the Timer_A UART
    //------------------------------------------------------------------------------
    void TimerA_UART_tx(unsigned char byte)
    {
        while (TACCTL0 & CCIE);                 // Ensure last char got TX'd
        TACCR0 = TAR;                           // Current state of TA counter
        TACCR0 += UART_TBIT;                    // One bit time till first bit
        TACCTL0 = OUTMOD0 + CCIE;               // Set TXD on EQU0, Int
        txData = byte;                          // Load global variable
        txData |= 0x100;                        // Add mark stop bit to TXData
        txData <<= 1;                           // Add space start bit
    }

    //------------------------------------------------------------------------------
    // Prints a string over using the Timer_A UART
    //------------------------------------------------------------------------------
    void TimerA_UART_print(char *string)
    {
        while (*string) {
            TimerA_UART_tx(*string++);
        }
    }
    //------------------------------------------------------------------------------
    // Timer_A UART - Transmit Interrupt Handler
    //------------------------------------------------------------------------------
    #pragma vector = TIMERA0_VECTOR
    __interrupt void Timer_A0_ISR(void)
    {
        static unsigned char txBitCnt = 10;

        TACCR0 += UART_TBIT;                    // Add Offset to CCRx
        if (txBitCnt == 0) {                    // All bits TXed?
            TACCTL0 &= ~CCIE;                   // All bits TXed, disable interrupt
            txBitCnt = 10;                      // Re-load bit counter
        }
        else {
            if (txData & 0x01) {
              TACCTL0 &= ~OUTMOD2;              // TX Mark '1'
            }
            else {
              TACCTL0 |= OUTMOD2;               // TX Space '0'
            }
            txData >>= 1;
            txBitCnt--;
        }
    }
    //------------------------------------------------------------------------------
    // Timer_A UART - Receive Interrupt Handler
    //------------------------------------------------------------------------------
    #pragma vector = TIMERA1_VECTOR
    __interrupt void Timer_A1_ISR(void)
    {
        static unsigned char rxBitCnt = 8;
        static unsigned char rxData = 0;

        switch (__even_in_range(TAIV, TAIV_TAIFG)) { // Use calculated branching
            case TAIV_TACCR1:                        // TACCR1 CCIFG - UART RX
                TACCR1 += UART_TBIT;                 // Add Offset to CCRx
                if (TACCTL1 & CAP) {                 // Capture mode = start bit edge
                    TACCTL1 &= ~CAP;                 // Switch capture to compare mode
                    TACCR1 += UART_TBIT_DIV_2;       // Point CCRx to middle of D0
                }
                else {
                    rxData >>= 1;
                    if (TACCTL1 & SCCI) {            // Get bit waiting in receive latch
                        rxData |= 0x80;
                    }
                    rxBitCnt--;
                    if (rxBitCnt == 0) {             // All bits RXed?
                        rxBuffer = rxData;           // Store in global variable
                        rxBitCnt = 8;                // Re-load bit counter
                        TACCTL1 |= CAP;              // Switch compare to capture mode
                        __bic_SR_register_on_exit(LPM0_bits);  // Clear LPM0 bits from 0(SR)
                    }
                }
                break;
        }
    }

    //------------------------------------------------------------------------------
